﻿



local DSxReusableLib_MAJOR, DSxReusableLib_MINOR = "DSxReusableLib", 21
local DSxReusableLib, oldminor = LibStub:NewLibrary(DSxReusableLib_MAJOR, DSxReusableLib_MINOR)

if not DSxReusableLib then return end -- No upgrade needed

--
-- Private Section
--

--hash of existing DSxReusableLibObjects
local DSxReusableLibObjHash = {}

--essentially a deep copy
local function init(tbl, defaults)
    for k,v in pairs(defaults) do
        if type(v) == "table" then
            tbl[k]={}
            init(tbl[k],v)
        else
            tbl[k]=v
        end
    end
end

--deep copy preserving existing tables, but cleaning out any extra settings on the tbl
local function clean(tbl, defaults)
    --much like init
    for k,v in pairs(defaults) do
        if type(v) == "table" then
            if not tbl[k] or type(tbl[k]) ~= "table" then
                tbl[k]={}
            end
            clean(tbl[k],v)
        else
            tbl[k]=v
        end
    end
    
    --this is the cleaning step
    for k,v in pairs(tbl) do
        --we have something set on tbl that didn't exist in the default so chuck it
        if not defaults[k] then
            tbl[k]=nil
        end
    end
end

--
-- Public Section
--

local methods = {
    ["Borrow"] = function(self)
        if not DSxReusableLibObjHash[self] then
            error("DSxReusableObject:Borrow() called on non-DSxReusableObject", 2)
        end
        
        local s = #self.storage
        local ret
        if s > 0 then
            ret = self.storage[s]
            table.remove(self.storage,s)
        else
            ret = {}
            init(ret, self.defaults)
        end
        
        return ret
    end,
    
    ["Return"] = function(self, val)
        if not DSxReusableLibObjHash[self] then
            error("DSxReusableObject:Return() called on non-DSxReusableObject", 2)
        end
        
        if not val or type(val) ~= "table" then 
            error("Usage: DSxReusableObject:Return(val): 'val' - table expected.", 2)
        end
        
        clean(val, self.defaults)
        table.insert(self.storage,val)
    end,
}

function DSxReusableLib:New(defaults)
    if type(defaults) ~= "table" then
		error("Usage: DSxReusableLib:New(defaults): 'defaults' - table expected.", 2)
	end
    
    local reusable = {
        defaults=defaults,
        storage={}
    }
    
    for method, func in pairs(methods) do
		reusable[method] = func
	end
    
    DSxReusableLibObjHash[reusable]=1
    
    return reusable
end